program AutoInitDMASoundRecording;
{****************************************************************************
** Demonstration of recording a WAV file through the Sound Blaster         **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************}

uses Dos, Crt;

const
  CD = 2;
  Microphone = 4;
  LineIn = 5;

{The simplest header a WAV file can have is this:}
type SimpleHeader = record
  RIFF : LongInt;
  NextChunkSize : LongInt;
  WAVE : LongInt;
  fmt : LongInt;
  fmtDataLength : LongInt;
  WaveType : Word;
  Channels : Word;
  SampleRate : LongInt;
  BytesPerSecond : LongInt;
  BlockAlignment : Word;
  BitResolution : Word;
  data : LongInt;
  SoundLength : LongInt;
end;

var
  Base      : Word;    {Sound Blaster base address}
  DMA       : Byte;    {The DMA channel}
  IRQ       : Byte;    {The IRQ level}
  OldIRQ    : Pointer; {Pointer to old interrupt routines}

  WBuffer   : Word;    {Write buffer indicator}
  DMABuffer : Pointer; {Pointer to DMA buffer}

  WAVFile   : File;    {Handle of the WAV file}
  Header    : SimpleHeader;
  Recorded  : LongInt; {Amount of samples recorded}

{****************************************************************************
** Checks to see if a Sound Blaster exists at a given address, returns     **
** true if Sound Blaster found, false if not.                              **
****************************************************************************}
function ResetDSP(Test : Word) : Boolean;
begin
  {Reset the DSP}
  Port[Test + $6] := 1;
  Delay(10);
  Port[Test + $6] := 0;
  Delay(10);
  {Check if reset was succesfull}
  if (Port[Test + $E] and $80 = $80) and (Port[Test + $A] = $AA) then begin
    {DSP was found}
    ResetDSP := true;
    Base := Test;
  end else
    {No DSP was found}
    ResetDSP := false;
end;

{****************************************************************************
** Send a byte to the DSP (Digital Signal Processor) on the Sound Blaster  **
****************************************************************************}
procedure WriteDSP(Value : byte);
begin
  {Wait for the DSP to be ready to accept data}
  while Port[Base + $C] and $80 <> 0 do;
  {Send byte}
  Port[Base + $C] := value;
end;

{****************************************************************************
** Writes a value to the SB's mixer chip                                   **
****************************************************************************}
procedure WriteMixer (R, V : Byte);
begin
  Port [Base + 4] := R;
  Port [Base + 5] := V;
end;

{****************************************************************************
** Selects the source device for recording.                                **
** Valid values are: Microphone, CD and LineIn                             **
****************************************************************************}
procedure SelectSource (Source : Byte);
begin
  {For older SB's}
  WriteMixer ($C, Source);
  {For newer SB's}
  begin
    {Disable right channel}
    WriteMixer ($3E, 0);
    {Record all sound through left channel}
    case Source of
      Microphone : WriteMixer ($3D,  1);
      CD         : WriteMixer ($3D,  6);
      LineIn     : WriteMixer ($3D, 24);
    end;
    {Set volume of whatever is recording to full}
    case Source of
      Microphone : WriteMixer ($0A,   7);
      CD         : WriteMixer ($28, 255);
      LineIn     : WriteMixer ($2E, 255);
    end;
    {Allow line and CD to be heard through speakers, but not microphone}
    case Source of
      Microphone : WriteMixer ($3C,  0);
      CD         : WriteMixer ($3C,  6);
      LineIn     : WriteMixer ($3C, 24);
    end;
  end;
end;

{****************************************************************************
** Asks the user for a filename and prepares that file for recording.      **
****************************************************************************}
procedure PrepareFile;
var
  FileName : String;

begin
  {Ask the user for a file name}
  Write ('Please give filename       : ');
  Readln (FileName);
  {Open the file for output}
  Assign (WAVFile, FileName);
  Rewrite (WAVFile, 1);
  {Save a dummy header}
  BlockWrite (WAVFile, Header, SizeOf (Header));
end;

{****************************************************************************
** Starts recording. The DMA controller is programmed with a block length  **
** of 32K - the entire buffer. The DSP is instructed to play blocks of 16K **
** and then generate an interrupt (which allows the program to save the    **
** parts that have already been recorded)                                  **
****************************************************************************}
procedure StartRecording;
var
  LinearAddress : LongInt;
  Page, OffSet  : Word;

begin
  {Nothing recorded yet}
  Recorded := 0;

  WriteDSP($D3);             {DSP-command D3h - Disable speaker, required
                             on older SB's}
  WriteDSP($40);             {DSP-command 40h - Set sample frequency}
  WriteDSP(165);             {Write time constant for 11025Hz}
  {
    The time constant is calculated as follows:
       (65536 - (256000000 div frequency)) shr 8
  }

  {Convert pointer to linear address}
  LinearAddress := Seg (DMABuffer^);
  LinearAddress := LinearAddress shl 4 + Ofs (DMABuffer^);
  Page := LinearAddress shr 16;  {Calculate page}
  OffSet := LinearAddress and $FFFF; {Calculate offset in the page}
  Port[$0A] := 4 or DMA;     {Mask DMA channel}
  Port[$0C] := 0;            {Clear byte pointer}
  Port[$0B] := $54 or DMA;   {Set mode}
  {
    The mode consists of the following:
    $54+x = binary 01 01 01 xx
                   |  |  |  |
                   |  |  |  +- DMA channel
                   |  |  +---- Write operation (the DSP writes to memory)
                   |  +------- Auto init mode
                   +---------- Block mode
  }
  Port[DMA shl 1] := Lo(OffSet);   {Write the offset to the DMA controller}
  Port[DMA shl 1] := Hi(OffSet);

  case DMA of
    0 : Port[$87] := Page;             {Write the page to the DMA controller}
    1 : Port[$83] := Page;             {Write the page to the DMA controller}
    3 : Port[$82] := Page;             {Write the page to the DMA controller}
  end;

  Port[DMA shl 1 + 1] := $FF;          {Set the block length to $7FFF = 32 Kbyte}
  Port[DMA shl 1 + 1] := $7F;

  Port[$0A] := DMA;          {Unmask DMA channel}

  WriteDSP($48);             {DSP-command 48h - Set block length}
  WriteDSP($FF);             {Set the block length to $3FFF = 16 Kbyte}
  WriteDSP($3F);
  WriteDSP($2C);             {DSP-command 2Ch - Start auto-init recording}
end;

{****************************************************************************
** Saves one half of the DMA buffer to the file                            **
****************************************************************************}
procedure SaveBuffer (Buffer : Word);
begin
  BlockWrite (WAVFile, Mem [Seg(DMABuffer^):Ofs(DMABuffer^) + Buffer shl 14], 16384);
  inc (Recorded, 16384);
end;

{****************************************************************************
** Saves the final part of the buffer and finalizes the WAV file           **
****************************************************************************}
procedure EndRecording;
var
  Counter, OffSet, DMAPointer : Word;

begin
  {Stops DMA-transfer}
  WriteDSP ($D0);
  WriteDSP ($DA);

  {Read DMA pointer from DMA controller}
  DMAPointer := Port [1 + DMA shl 1];
  DMAPointer := DMAPointer + Port [1 + DMA shl 1] shl 8;

  {
    DMAPointer contains amount that remains to be played.
    This is converted to the offset of the current sample
  }
  OffSet := $7FFF - DMAPointer;

  {Save the last bit}
  BlockWrite (WAVFile,
              Mem [Seg(DMABuffer^):Ofs(DMABuffer^) + OffSet and 16384],
              OffSet and 16383);

  {The total amount recorded is now available}
  inc (Recorded, OffSet and 16383);

  {Save the actual header}
  Header.RIFF := $46464952;              {"RIFF"}
  Header.NextChunkSize := Recorded + 36; {Size of the WAVE chunk}
  Header.WAVE := $45564157;              {"WAVE"}
  Header.fmt := $20746D66;               {"fmt "}
  Header.fmtDataLength := 16;            {length of fmt chunk}
  Header.WaveType := 1;                  {Standard uncompressed WAV file}
  Header.Channels := 1;                  {Mono}
  Header.SampleRate := 11025;            {11KHz}
  Header.BytesPerSecond := Header.Channels * Header.SampleRate;
  Header.BlockAlignment := 1;            {1 for 8bit, 2 for 16bit}
  Header.BitResolution := 8;             {8bit WAV file}
  Header.data := $61746164;              {"data"}
  Header.SoundLength := Recorded;
  {Save the actual header}
  Seek (WAVFile, 0);
  BlockWrite (WAVFile, Header, SizeOf (Header));
  Close (WAVFile);
end;

{****************************************************************************
** IRQ service routine - this is called when the DSP has finished playing  **
** a block                                                                 **
****************************************************************************}
procedure ServiceIRQ; interrupt;
var
  Temp : Byte;

begin
  {Relieve DSP}
  Temp := Port[$22E];
  {Acknowledge hardware interrupt}
  Port [$20] := $20;
  {Acknowledge cascade interrupt for IRQ 2 and 10}
  if IRQ in [2, 10] then Port [$A0] := $20;
  {Save buffer}
  SaveBuffer (WBuffer);
  {Toggle between write buffers}
  WBuffer := WBuffer xor 1;
end;

{****************************************************************************
** This procedure allocates 32K of memory to the DMA buffer and makes sure **
** that no page boundary is crossed                                        **
****************************************************************************}
procedure AssignBuffer;
var
  TempBuf       : Pointer; {Temporary pointer}
  LinearAddress : LongInt;
  Page1, Page2  : Word;

begin
  {Assign 32K of memory}
  GetMem (TempBuf, 32768);

  {Calculate linear address}
  LinearAddress := Seg (TempBuf^);
  LinearAddress := LinearAddress shl 4 + Ofs (TempBuf^);
  {Calculate page at start of buffer}
  Page1 := LinearAddress shr 16;
  {Calculate page at end of buffer}
  Page2 := (LinearAddress + 32767) shr 16;

  {Check to see if a page boundary is crossed}
  if (Page1 <> Page2) then begin
    {If so, assign another part of memory to the buffer}
    GetMem (DMABuffer, 32768);
    FreeMem (TempBuf, 32768);
  end else begin
    {otherwise, use the part we've already allocated}
    DMABuffer := TempBuf;
  end;
end;

{****************************************************************************
** This procedure checks the possible addresses to see whether a Sound     **
** Blaster is installed.                                                   **
****************************************************************************}
procedure FindSB;
var
  Temp : Byte;

begin
  {Check for Sound Blaster, address: ports 220, 230, 240, 250, 260 or 280}
  for Temp := 1 to 8 do begin
    if Temp <> 7 then
    if ResetDSP ($200 + Temp shl 4) then Break;
  end;
  if Temp = 9 then begin
    {or none at all}
    Writeln ('No Sound Blaster found');
    Halt;
  end else Writeln ('Sound Blaster found at 2', Temp, '0h');
end;

{****************************************************************************
** This procedure asks the user to input the DMA channel and IRQ level and **
** sets up the program accordingly.                                        **
****************************************************************************}
procedure InitIRQandDMA;
begin
  {Ask for IRQ and DMA}
  Write ('Please specify DMA channel : '); Read (DMA);
  Write ('Please specify IRQ level   : '); Read (IRQ);

  {Save old IRQ vector}
  case IRQ of
    2 : GetIntVec($71, OldIRQ);
   10 : GetIntVec($72, OldIRQ);
   11 : GetIntVec($73, OldIRQ);
  else
    GetIntVec (8 + IRQ, OldIRQ);
  end;
  {Set new IRQ vector}
  case IRQ of
    2 : SetIntVec($71, Addr (ServiceIRQ));
   10 : SetIntVec($72, Addr (ServiceIRQ));
   11 : SetIntVec($73, Addr (ServiceIRQ));
  else
    SetIntVec (8 + IRQ, Addr (ServiceIRQ));
  end;
  {Enable IRQ}
  case IRQ of
    2 : Port[$A1] := Port[$A1] and not 2;
   10 : Port[$A1] := Port[$A1] and not 4;
   11 : Port[$A1] := Port[$A1] and not 8;
  else
    Port[$21] := Port[$21] and not (1 shl IRQ);
  end;
  if IRQ in [2, 10, 11] then Port[$21] := Port[$21] and not 4;
end;

{****************************************************************************
** This procedure releases the DMA channel and IRQ level                   **
****************************************************************************}
procedure ReleaseIRQandDMA;
begin
  {Free interrupt vectors used to service IRQs}
  case IRQ of
    2 : SetIntVec($71, OldIRQ);
   10 : SetIntVec($72, OldIRQ);
   11 : SetIntVec($73, OldIRQ);
  else
    SetIntVec (8 + IRQ, OldIRQ);
  end;

  {Mask IRQs}
  case IRQ of
    2 : Port[$A1] := Port[$A1] or 2;
   10 : Port[$A1] := Port[$A1] or 4;
   11 : Port[$A1] := Port[$A1] or 8;
  else
    Port[$21] := Port[$21] or (1 shl IRQ);
  end;
  if IRQ in [2, 10, 11] then Port[$21] := Port[$21] or 4;
end;

begin
  {Clear screen}
  ClrScr;
  Writeln ('Demonstration of recording samples using DMA - by Steven Don');
  Writeln ('------------------------------------------------------------');

  {Look for the Sound Blaster}
  FindSB;

  {Assign memory to the DMA Buffer}
  AssignBuffer;

  {Set write buffer to first buffer}
  WBuffer := 0;

  {Select microphone as the source}
  SelectSource (Microphone);

  {Ask the user for a filename and prepare the file for output}
  PrepareFile;

  {Ask the user for IRQ and DMA channels}
  InitIRQandDMA;

  {Start recording}
  StartRecording;

  {Wait until escape is pressed}
  repeat until ReadKey = Chr(27);

  {End the recording}
  EndRecording;

  {Free the memory allocated to the sound buffer}
  FreeMem (DMABuffer, 32768);

  {Release the IRQ and DMA channels}
  ReleaseIRQandDMA;
end.